#include <../src/kernel/objbase.h>
#include <../src/adt/tstack.h>

/*
 *	Implementation of u8_stack_t and vp_stack_t.
 */
TYPED_STACK_IMPLEMENT_CLASS(u8,u8_t,u8_p)
TYPED_STACK_IMPLEMENT_CLASS(vp,void_p,void_pp)

/*
 * Additional methods for 'stack of pointers'
 */

/*
 *	Finds given pointer, returns 'true' if found.
 */
vboolean	V_GLOBAL
vpsFind(vp_stack_t *st,void *ptr,u32_t *pos_return)
{
	u32_t i,len;
	void **objects;
	V_ASSERT(st);
	len=vpsCount(st);
	objects=(void**)vpsGetStorage(st);
	V_ASSERT(len==0 || objects);
	for(i=0;i<len;i++)
		if(objects[i]==ptr)
		{
			if(pos_return)
				*pos_return=i;
			return true;
		}
	return false;
}

/*
 *	Stores pointer at the first non-NULL position, returns position.
 */
u32_t	V_GLOBAL
vpsStore(vp_stack_t *st,void *ptr)
{
	u32_t i,len;
	void **objects;
	V_ASSERT(st);
	len=vpsCount(st);
	objects=(void**)vpsGetStorage(st);
	V_ASSERT(len==0 || objects);
	for(i=0;i<len;i++)
		if(!objects[i])
		{
			objects[i]=ptr;
			return i;
		}
	vpsPush(st,ptr);
	return len;
}

/*
 * Recursive sorting procedure for vpsSort()
 */
V_LOCAL void
vpsSort_r(void **pointers,int l,int r,compar_proc_t compar,void *param)
{
	int i,j,m;
	void *x,*y;
	i = l;
	j = r;
	m = (l+r) >> 1;
	x = pointers[ m ];
	do
	{
		while(compar(param,pointers[i],x)<0)
			i++;
		while(compar(param,x,pointers[j])<0)
			j--;
		if(i<=j)
		{
			y=pointers[i]; 
			pointers[i]=pointers[j]; 
			pointers[j]=y;
			i++; 
			j--;
		}
	} while(i<=j);
	if(l<j)
		vpsSort_r(pointers,l,j,compar,param);
	if(i<r)
		vpsSort_r(pointers,i,r,compar,param);
}

/*
 * Thread-safe version of qsort().
 */
void	V_GLOBAL
vpsSort(vp_stack_t *objects,compar_proc_t compar,void *param)
{
	u32_t len;
	void *check_obj;
	if(!objects || !compar)
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,"vpsSort(%X,%X,%X): incorrect parameter(s).\n",
			objects,compar,param);
#endif	/* CHECKED_BUILD */
		return;
	}
	len=vpsCount(objects);
    if(len<2)
    {
#ifdef	CHECKED_BUILD
		conlPuts(LOG_TRACE,"vpsSort(): nothing to sort.\n");
#endif	/* CHECKED_BUILD */
		return;
	}
	if(len>=(MAXLONG>>1))
	{
#ifdef	CHECKED_BUILD
		conlPrintf(LOG_ERROR,
			"vpsSort(): too many objects (%u).\n",len);
#endif	/* CHECKED_BUILD */
		return;
	}
	/*
     * Comparison procedure SHOULD return zero for equal objects,
	 * otherwise we get core dump; so test intelligence of 'compar'.
	 */
	check_obj=vpsGetStorage(objects)[0];
	if(compar(param,check_obj,check_obj)) /* idential objects ! */
	{
#ifdef	CHECKED_BUILD
		conlPuts(LOG_ERROR,
			"vpsSort(): brain-dead comparison procedure.\n");
#endif	/* CHECKED_BUILD */
		return;
	}
    /* Execute qsort algorithm */
	vpsSort_r(vpsGetStorage(objects),0,(int)(len-1),compar,param);
    return;
}

/*
 * Implementation of "Stack of blocks"
 */

void	V_GLOBAL
stackInit(bstack_t *stack,unsigned element_size)
{
	V_ASSERT(stack && element_size);
	stack->storage = stack->ptr = stack->endptr = NULL;
	stack->element_size = element_size;
    stack->capacity ?
}

void	V_GLOBAL
stackClear(bstack_t *stack)
{
	V_ASSERT(stack);
	if(stack->storage)
	{
		memFree(stack->storage);
		stack->storage = NULL;
	}
	stack->ptr = stack->endptr = NULL;
    stack->capacity ?
}

void	V_GLOBAL
stackPush(bstack_t *stack,void *element)
{
	V_ASSERT(stack && element);
	if(stack->ptr == stack->endptr)
	{
		/* Allocate new storage */
		unsigned nextsize, byte_diff, byte_len;
        byte_diff = stack->ptr - stack->storage;
		GREATER_POWER_TWO(nextsize,stack->capacity,unsigned);
        byte_len = nextsize * stack->element_size;
		stack->storage=(u8_p)memRealloc(stack->storage,byte_len);
		stack->ptr = stack->storage + byte_diff;
        stack->endptr = stack->storage + byte_len;
		stack->capacity = nextsize;
	}
	memCopy(stack->ptr, element, stack->element_size);
	stack->ptr += stack->element_size;
}

int		V_GLOBAL
stackPop(bstack_t *stack,void *element)
{
	V_ASSERT(stack && element);
	if(stack->ptr > stack->storage)
	{
		stack->ptr -= stack->element_size;
		memCopy(element,stack->ptr,stack->element_size);
		return 1;
	}
	return 0;
}

int		V_GLOBAL
stackPopBuffer(bstack_t *stack,void *dst,unsigned num_elements)
{
	V_ASSERT(stack && dst && num_elements);
	if(num_elements>stack->tos_cursor)
		num_elements=stack->tos_cursor;
	stack->tos_cursor-=num_elements;
	stack->ptr-=num_elements*stack->element_size;
	memCopy(dst,stack->ptr,num_elements*stack->element_size);
	return num_elements;
}

void	V_GLOBAL
stackPushBuffer(bstack_t *stack,void *buffer,unsigned num_elements)
{
	unsigned total;
	V_ASSERT(stack && buffer && num_elements);
	total=stack->tos_cursor+num_elements;
	if(total > stack->capacity)
	{
		NOT_LESS_POWER_TWO(total,total,unsigned);
		stack->storage=(unsigned char *)memRealloc(stack->storage,
			total*stack->element_size);
		stack->capacity=total;
		stack->ptr=stack->storage + stack->tos_cursor * stack->element_size;
	}
	memCopy(stack->ptr,buffer,num_elements*stack->element_size);
	stack->ptr+=num_elements*stack->element_size;
	stack->tos_cursor+=num_elements;
}

void	V_GLOBAL
stackResize(bstack_t *stack,unsigned newsize)
{
	V_ASSERT(stack);
	if(newsize > stack->capacity)
	{
    	/* We must allocate new storage */
    	unsigned nextlen;
    	NOT_LESS_POWER_TWO(nextlen,newsize,unsigned);
    	stack->storage=(unsigned char*)memRealloc(stack->storage,
    		nextlen*stack->element_size);
    	stack->capacity=nextlen;
		stack->ptr=stack->storage + stack->tos_cursor * stack->element_size;
    	/* Fill rest with zeros */
    	memZero(stack->ptr,(newsize - stack->tos_cursor) * stack->element_size);
    }
    else
    {
		if(newsize > stack->tos_cursor)
    		memZero(stack->ptr,(newsize-stack->tos_cursor)*stack->element_size);
    }
	stack->tos_cursor=newsize;
	stack->ptr=stack->storage + stack->tos_cursor * stack->element_size;
}

void	V_GLOBAL
stackAdjust(bstack_t *stack)
{
	V_ASSERT(stack);
	if(stack->tos_cursor)
	{
    	stack->storage=(unsigned char*)memRealloc(stack->storage,
    		stack->tos_cursor*stack->element_size);
    	stack->capacity=stack->tos_cursor;
    }
    else
    {
		memFree(stack->storage);
        stack->storage=NULL;
	}
	stack->capacity=stack->tos_cursor;
	stack->ptr=stack->storage + stack->tos_cursor * stack->element_size;
}

void	V_GLOBAL
stackSet(bstack_t *stack,unsigned pos,void *element)
{
	V_ASSERT(stack && element);
	if(pos>stack->capacity)
		stackResize(stack,pos);
	stackPush(stack,element);
}

void	V_GLOBAL
stackConcat(bstack_t *dst,bstack_t *src)
{
	V_ASSERT(dst && src && dst->element_size == src->element_size);
	stackPushBuffer(dst,src->storage,src->tos_cursor);
}

#ifdef		STACK_NO_MACRO

unsigned	V_GLOBAL
stackCount(bstack_t *stack)
{
	return stack->tos_cursor;
}

void	V_GLOBAL
stackGet(bstack_t *stack,void *dst,unsigned pos)
{
	memCopy(dst,(u8_p)stack->storage + pos * stack->element_size,
		stack->element_size);
}

void_p	V_GLOBAL
stackGetStorage(bstack_t *stack)
{
	return (stack)->storage;
}

#endif		/* STACK_NO_MACRO */

/*EOF*/
