#ifndef	TSTACK_H
#define TSTACK_H

/*
 * Templates for stacks.
 * Note: Also any LibV client can safely include this header to produce
 * code for new stacks, I don't recommend to make this header visible:
 * it can significantly increase compile time.
 */

/*
 * Internal macros for "typed stacks"
 */

/*
 * Sets all fields of stack to zeros.
 * Must be called before any other operations.
 */
#define	TYPED_STACK_INIT(stack) \
		(stack).capacity = (stack).tos_cursor=0; \
		(stack).elements = NULL;

/*
 * Appends next element to TOS.
 */
#define	TYPED_STACK_PUSH(stack,element,type_elem) \
		if((stack).tos_cursor >= (stack).capacity) { \
			/* Not enought space: realloc() storage */ \
			unsigned nextsize; \
			GREATER_POWER_TWO(nextsize,(stack).tos_cursor,unsigned); \
			(stack).elements=(type_elem*)memRealloc((stack).elements, \
            	nextsize*sizeof(type_elem)); \
			(stack).capacity=nextsize; \
		} \
		(stack).elements[(stack).tos_cursor++]=element;

/*
 * Appends 'num_elements' to stack.
 */
#define	TYPED_STACK_PUSH_BUFFER(stack,buffer,num_elements,type_elem) { \
			unsigned total; \
			total=(stack).tos_cursor + (num_elements); \
			if(total > (stack).capacity) { \
				NOT_LESS_POWER_TWO(total,total,unsigned); \
				(stack).elements=(type_elem*)memRealloc( \
                	(stack).elements,total*sizeof(type_elem)); \
				(stack).capacity=total; \
			} \
			memCopy((stack).elements+(stack).tos_cursor,(buffer), \
				(num_elements)*sizeof(type_elem)); \
			(stack).tos_cursor+=(num_elements); \
		}

/*
 * Copies memory from stack into 'dst' in FIFO order.
 */
#define	TYPED_STACK_POP_BUFFER_AS_FIFO(stack,dst,num_elements,type) \
	if((num_elements) > (stack).tos_cursor) \
		(num_elements)=(stack).tos_cursor; \
	(stack).tos_cursor-=(num_elements); \
	memCopy((dst), \
    	(stack).elements+(stack).tos_cursor,(num_elements)*sizeof(type)); \

/*
 * Sets TOS cursor to new position.
 * Buffer will be resized to greater size, if necessary.
 */
#define	TYPED_STACK_RESIZE(stack,newsize,type_elem) \
		if((newsize) > (stack).capacity) { \
			/* Allocate new storage */ \
			unsigned nextlen; \
			NOT_LESS_POWER_TWO(nextlen,(newsize),unsigned); \
			(stack).elements=(type_elem*)memRealloc((stack).elements, \
            	nextlen*sizeof(type_elem)); \
			/* Fill tail with zeros */ \
			memZero((stack).elements+(stack).tos_cursor, \
				((newsize)-(stack).tos_cursor)*sizeof(type_elem)); \
			(stack).capacity=nextlen; \
		} else { \
			/* We fit in current storage */ \
			if((newsize) > (stack).tos_cursor) \
				memZero((stack).elements+(stack).tos_cursor, \
					((newsize)-(stack).tos_cursor)*sizeof(type_elem)); \
		} \
		(stack).tos_cursor=(newsize);

/*
 * Frees unused memory.
 */
#define	TYPED_STACK_FREE_UNUSED_MEMORY(stack,type_elem) \
		if((stack).tos_cursor) \
			/* We must memRealloc() to lower size */ \
			(stack).elements=(type_elem*)memRealloc((stack).elements, \
            	(stack).tos_cursor*sizeof(type_elem));\
		else { \
        	/* There are no data - clear stack */ \
			memFree((stack).elements); \
            (stack).elements=NULL; \
		} \
		(stack).capacity=(stack).tos_cursor;\

/*	----------------------------------------------------------------------
 *	Templates for "typed stacks"
 */
#define	TYPED_STACK_DEFINE_INIT(type_prefix) \
void	V_GLOBAL \
type_prefix##sInit(type_prefix##_stack_t *stack) \
{ \
	V_ASSERT(stack); \
	TYPED_STACK_INIT(*stack); \
}

#define	TYPED_STACK_DEFINE_CLEAR(type_prefix) \
void	V_GLOBAL \
type_prefix##sClear(type_prefix##_stack_t *stack) \
{ \
	V_ASSERT(stack); \
	if(stack->buf) \
		memFree(stack->buf); \
	TYPED_STACK_INIT(*stack); \
}

#define	TYPED_STACK_DEFINE_PUSH(type_prefix,type) \
void	V_GLOBAL \
type_prefix##sPush(type_prefix##_stack_t *stack,type element) \
{ \
	V_ASSERT(stack); \
	TYPED_STACK_PUSH(*stack,element,type); \
}

#define	TYPED_STACK_DEFINE_POP(type_prefix,type) \
int		V_GLOBAL \
type_prefix##sPop(type_prefix##_stack_t *stack,type *element) \
{ \
	V_ASSERT(stack && element); \
	if(stack->next_elem) { \
		*element=stack->buf[--stack->next_elem]; \
		return 1; \
	} else return 0; \
}

#define	TYPED_STACK_DEFINE_PUSH_BUFFER(type_prefix,type) \
void	V_GLOBAL \
type_prefix##sPushBuffer(type_prefix##_stack_t *stack, \
						type *src,unsigned num_elements) \
{ \
	V_ASSERT(stack && src && num_elements); \
	TYPED_STACK_PUSH_BUFFER(*stack,src,num_elements,type); \
}

#define	TYPED_STACK_DEFINE_POP_BUFFER(type_prefix,type) \
int		V_GLOBAL \
type_prefix##sPopBuffer(type_prefix##_stack_t *stack, \
						type *dst,unsigned num_elements) \
{ \
	V_ASSERT(stack && dst && num_elements); \
    TYPED_STACK_POP_BUFFER_AS_FIFO(*stack,dst,num_elements,type); \
	return num_elements; \
}

#define	TYPED_STACK_DEFINE_RESIZE(type_prefix,type) \
void	V_GLOBAL \
type_prefix##sResize(type_prefix##_stack_t *stack,unsigned newsize) \
{ \
	V_ASSERT(stack); \
	TYPED_STACK_RESIZE(*stack,newsize,type); \
}

#define	TYPED_STACK_DEFINE_ADJUST(type_prefix,type) \
void	V_GLOBAL \
type_prefix##sAdjust(type_prefix##_stack_t *stack) \
{ \
	V_ASSERT(stack); \
	TYPED_STACK_FREE_UNUSED_MEMORY(*stack,type); \
}

#define	TYPED_STACK_DEFINE_COUNT(type_prefix) \
unsigned	V_GLOBAL \
type_prefix##sCount(type_prefix##_stack_t *stack) \
{ \
	V_ASSERT(stack); \
	return stack->next_elem; \
}

#define	TYPED_STACK_DEFINE_GET(type_prefix,type) \
type	V_GLOBAL \
type_prefix##sGet(type_prefix##_stack_t *stack,unsigned pos) \
{ \
	V_ASSERT(stack && stack->n_elem > pos); \
	return stack->buf[pos]; \
}

#define	TYPED_STACK_DEFINE_GET_STORAGE(type_prefix,type_ptr) \
type_ptr	V_GLOBAL \
type_prefix##sGetStorage(type_prefix##_stack_t *stack) \
{ \
	V_ASSERT(stack); \
	return stack->buf; \
}

#define	TYPED_STACK_DEFINE_SET(type_prefix,type) \
void	V_GLOBAL \
type_prefix##sSet(type_prefix##_stack_t *stack,unsigned pos,type elemnt) \
{ \
	V_ASSERT(stack); \
	if(pos >= stack->n_elem) \
	{ \
		TYPED_STACK_RESIZE(*stack,pos+1,type); \
	} \
	stack->buf[pos]=elemnt; \
}

#define	TYPED_STACK_DEFINE_CONCAT(type_prefix,type) \
void	V_GLOBAL \
type_prefix##sConcat(type_prefix##_stack_t *dst, \
					 type_prefix##_stack_t *src) \
{ \
	V_ASSERT(src && dst); \
	TYPED_STACK_PUSH_BUFFER(*dst,src->buf,src->next_elem,type); \
}

/*	Now collect all together ------------------------------------------ */
/*  To define new stack: invoke 'TYPED_STACK_IMPLEMENT_CLASS' in global
 *	scope, passing:
 *		- 'method_prefix' - letter(s) for prefix of routines;
 *		- 'element_type' - type of element (must be scalar or any other
 *			type with assignment operation defined);
 *		- 'pointer_type' - pointer to 'element_type'.
 */
#ifdef	STACK_NO_MACRO 
#define	TYPED_STACK_IMPLEMENT_CLASS(method_prefix, \
			element_type,pointer_type) \
		TYPED_STACK_DEFINE_INIT(method_prefix) \
		TYPED_STACK_DEFINE_CLEAR(method_prefix) \
		TYPED_STACK_DEFINE_PUSH(method_prefix,element_type) \
		TYPED_STACK_DEFINE_PUSH_BUFFER(method_prefix,element_type) \
		TYPED_STACK_DEFINE_POP(method_prefix,element_type) \
		TYPED_STACK_DEFINE_POP_BUFFER(method_prefix,element_type) \
		TYPED_STACK_DEFINE_RESIZE(method_prefix,element_type) \
		TYPED_STACK_DEFINE_ADJUST(method_prefix,element_type) \
		TYPED_STACK_DEFINE_CONCAT(method_prefix,element_type) \
		TYPED_STACK_DEFINE_SET(method_prefix,element_type) \
		TYPED_STACK_DEFINE_COUNT(method_prefix) \
		TYPED_STACK_DEFINE_GET(method_prefix,element_type) \
		TYPED_STACK_DEFINE_GET_STORAGE(method_prefix,pointer_type) 
#else	/* !STACK_NO_MACRO */
#define	TYPED_STACK_IMPLEMENT_CLASS(method_prefix, \
			element_type,pointer_type) \
		TYPED_STACK_DEFINE_INIT(method_prefix) \
		TYPED_STACK_DEFINE_CLEAR(method_prefix) \
		TYPED_STACK_DEFINE_PUSH(method_prefix,element_type) \
		TYPED_STACK_DEFINE_PUSH_BUFFER(method_prefix,element_type) \
		TYPED_STACK_DEFINE_POP(method_prefix,element_type) \
		TYPED_STACK_DEFINE_POP_BUFFER(method_prefix,element_type) \
		TYPED_STACK_DEFINE_RESIZE(method_prefix,element_type) \
		TYPED_STACK_DEFINE_ADJUST(method_prefix,element_type) \
		TYPED_STACK_DEFINE_CONCAT(method_prefix,element_type) \
		TYPED_STACK_DEFINE_SET(method_prefix,element_type) 
#endif	/* !STACK_NO_MACRO */

#endif	/* TSTACK_H */
